require 'rack/session/xfile'
require 'rack/mock'
require 'fileutils'
require 'tmpdir'


describe Rack::Session::XFile do

  def counter number
   %[{"counter"=>#{number}}]
  end

  def set_sid id
    if id =~ TheSession
      {'HTTP_COOKIE' => id}
    else
      {'HTTP_COOKIE' => "#{@session_key}=#{id}"}
    end
  end

  # Create an xfile out-of-band in the current session directory.
  def create_session data = nil
    app = Rack::Session::XFile.new(nil)
    sid = app.send(:generate_sid)
    app.write_session(EMPTY_REQUEST, sid, (data || {sid: sid}), {})
    sid
  end

  spec_tmp_dir  = Dir.mktmpdir 'xfile-specs-'
  @session_dir  = Rack::Session::XFile::DEFAULT_OPTIONS[:session_dir].
                  replace(File.join(spec_tmp_dir, 'xfile-sessions'))
  @session_key  = Rack::Session::XFile::DEFAULT_OPTIONS[:key]
  session_bits  = Rack::Session::XFile::DEFAULT_OPTIONS[:sidbits]
  session_chars = (session_bits / 8 / 3.0 * 4).ceil   # urlsafe_base64 length
  TheSession    = /^#{@session_key}=[0-9a-zA-Z]{#{session_chars}};/
  SESSION       = 'rack.session'.freeze
  SESSION_OPTS  = 'rack.session.options'.freeze
  EMPTY_SESSION = '{}'.freeze
  EMPTY_REQUEST = Rack::Request.new({}).freeze


  # Apps

  incrementor = lambda do |env|
    env[SESSION]['counter'] ||= 0
    env[SESSION]['counter'] += 1
    Rack::Response.new(env[SESSION].inspect).to_a
  end

  noop = lambda do |env|
    Rack::Response.new('noop').to_a
  end

  read_session_id = lambda do |env|
    Rack::Response.new(env[SESSION].id).to_a
  end

  read_session = lambda do |env|
    env[SESSION].send(:load_for_read!)
    Rack::Response.new(env[SESSION].inspect).to_a
  end

  defer_session = lambda do |env|
    env[SESSION_OPTS][:defer] = true
    incrementor.call(env)
  end

  drop_session = lambda do |env|
    env[SESSION_OPTS][:drop] = true
    incrementor.call(env)
  end

  renew_session = lambda do |env|
    env[SESSION_OPTS][:renew] = true
    incrementor.call(env)
  end


  before do
    @warnings = warnings = []
    Rack::Session::XFile.class_eval do
      define_method(:warn) { |m| warnings << m }
    end
  end

  after do
    FileUtils.rm_rf(@session_dir)
    Rack::Session::XFile.class_eval { remove_method :warn }
  end

  it 'sets default options' do
    key    = 'secret'
    dir    = File.join(spec_tmp_dir, 'secret-sessions')
    filter = /useragent/
    expire = 3600
    app    = Rack::Session::XFile.new(incrementor,
                                      key: key,
                                      session_dir: dir,
                                      user_agent_filter: filter,
                                      expire_after: expire)
    app.key.should.equal key
    app.default_options[:key].should.be.nil # deleted in superclass
    app.default_options[:session_dir].should.equal dir
    app.default_options[:user_agent_filter].should.equal filter
    app.default_options[:expire_after].should.equal expire
    cookie = Rack::MockRequest.new(app).get('/')['Set-Cookie']
    cookie.should.match /^#{key}=/
    FileUtils.rm_rf(dir)
  end

  it 'initializes session directory structure' do
    app = Rack::Session::XFile.new(noop)
    dir = app.default_options[:session_dir]
    Dir[File.join(dir, '**', '*')].count.should == # greedy match
      Rack::Session::XFile::SIDCharacters.count
    app.session_count.should.be.zero
  end

  it 'sets secure permissions when creating session directory' do
    # NOTE unable to test when no user/group ownership
    Rack::Session::XFile.new(noop)
    (File.stat(@session_dir).mode & 077).should.be.zero
    File.owned?(@session_dir).should.be.true
  end

  it 'raises exception if session directory is world accessible' do
    Dir.mkdir(@session_dir, 0707)
    should.raise(SecurityError) { Rack::Session::XFile.new(noop) }.
      message.should.include 'world accessible'
  end

  it 'raises exception if session directory is not user accessible' do
    Dir.mkdir @session_dir
    File.chmod 0000, @session_dir
    should.raise(Errno::EACCES) { Rack::Session::XFile.new(noop) }
    File.chmod 0100, @session_dir
    should.raise(Errno::EACCES) { Rack::Session::XFile.new(noop) }
    File.chmod 0200, @session_dir
    should.raise(Errno::EACCES) { Rack::Session::XFile.new(noop) }
    File.chmod 0400, @session_dir
    should.raise(Errno::EACCES) { Rack::Session::XFile.new(noop) }
    File.chmod 0700, @session_dir
    should.not.raise(Errno::EACCES) { Rack::Session::XFile.new(noop) }
  end

  it 'warns if session directory is group accessible but group owned' do
    Dir.mkdir(@session_dir, 0770)
    should.not.raise { Rack::Session::XFile.new(noop) }
    @warnings.count.should.equal 1
    @warnings.first.should.include 'group owned/accessible'
  end

  it 'returns false when writing session if cannot serialize data' do
    obj = Object.new
    def obj.singleton() nil end
    nonserializable = { data: obj }

    sid = create_session
    app = Rack::Session::XFile.new(noop)
    app.write_session(EMPTY_REQUEST, sid, nonserializable, {}).
      should.be.false
    @warnings.count.should.equal 1
    @warnings.first.should.include 'cannot write session: singleton'
  end

  it 'returns false when writing session if sid is nonexistent' do
    sid = 'nonexistent'
    app = Rack::Session::XFile.new(noop)
    app.write_session(EMPTY_REQUEST, sid, {}, {}).should.be.false
    @warnings.count.should.equal 1
    @warnings.first.should.include 'cannot write session: No such file'
  end

  it 'creates new xfile session' do
    app = Rack::Session::XFile.new(incrementor)
    res = Rack::MockRequest.new(app).get('/')
    res['Set-Cookie'].should.match TheSession
    res.body.should.equal counter(1)
  end

  it 'creates one xfile with six stateful requests' do
    app = Rack::Session::XFile.new(incrementor)
    req = Rack::MockRequest.new(app)

    app.session_count.should.be.zero
    cookie = req.get('/')['Set-Cookie']
    app.session_count.should.equal 1
    6.times { req.get('/', set_sid(cookie)) }
    app.session_count.should.equal 1
  end

  it 'creates six xfiles with six stateless requests' do
    app = Rack::Session::XFile.new(incrementor)
    req = Rack::MockRequest.new(app)

    app.session_count.should.be.zero
    6.times { req.get('/') }
    app.session_count.should.equal 6
  end

  it 'loads session data from existing xfile' do
    app = Rack::Session::XFile.new(read_session)
    sid = create_session(data = {'preloaded' => true})  # rack stringifies keys
    res = Rack::MockRequest.new(app).get('/', set_sid(sid))
    res.body.should.equal data.inspect
    app.session_count.should.equal 1
  end

  it 'merges session data from middleware in front' do
    app = Rack::Session::XFile.new(incrementor)
    res = Rack::MockRequest.new(app).get('/', SESSION => {foo: 'bar'})
    res.body.should.match /counter/
    res.body.should.match /foo/
  end

  it 'creates new session when xfile is nonexistent' do
    app = Rack::Session::XFile.new(read_session)
    res = Rack::MockRequest.new(app).get('/', set_sid('nonexistent'))
    res.body.should.equal EMPTY_SESSION
    res['Set-Cookie'].should.match TheSession
    app.session_count.should.equal 1
  end

  it 'creates new session when xfile is empty' do
    app = Rack::Session::XFile.new(read_session)
    sid = create_session
    xfile = app.send(:session_file, sid)
    File.truncate(xfile, 0)
    File.size(xfile).should.be.zero

    res = Rack::MockRequest.new(app).get('/', set_sid(sid))
    res.body.should.equal EMPTY_SESSION
    res['Set-Cookie'].should.be.nil # because sid already exists
    app.session_count.should.equal 1
    File.size(xfile).should.not.be.zero
  end

  it 'creates new session when xfile is corrupt' do
    app = Rack::Session::XFile.new(read_session)
    sid = create_session
    session_file = app.send(:session_file, sid)
    session_data = {'key' => 0}

    # return an empty session if marshaled session data is corrupt in any way
    critical_indices = [0, 3, 6]  # positions in marhaled +session_data+
    critical_indices.each do |i|  # that are suceptible to corruption
      corrupt_data = Marshal.dump(session_data).insert(i, 'FUBAR')
      File.write(session_file, corrupt_data)
      res = Rack::MockRequest.new(app).get('/', set_sid(sid))
      res.body.should.equal EMPTY_SESSION
      res['Set-Cookie'].should.be.nil
    end

    @warnings.count.should.equal critical_indices.count
    @warnings[0].should.include 'incompatible marshal file format' 
    @warnings[1].should.include 'dump format error for symbol' 
    @warnings[2].should.include 'marshal data too short' 
  end

  it 'protects against sid encoding and path traversal exploits' do
    app = Rack::Session::XFile.new(read_session)

    res1 = Rack::MockRequest.new(app).get('/', set_sid('%2e%2e%2f%2e%2e%2f%00'))
    res1.body.should.equal EMPTY_SESSION
    res1['Set-Cookie'].should.be.nil
    app.session_count.should.be.zero

    res2 = Rack::MockRequest.new(app).get('/', set_sid('../../../../etc/passwd'))
    res2.body.should.equal EMPTY_SESSION
    res2['Set-Cookie'].should.be.nil
    app.session_count.should.be.zero

    @warnings.size.should.equal 2
    @warnings[0].should.include 'tampered sid detected'
    @warnings[1].should.include 'tampered sid detected'
  end

  it 'returns new sid cookie with :renew option set' do
    inc_cntr  = Rack::Session::XFile.new(incrementor)
    read_sid  = Rack::Session::XFile.new(read_session_id) # doesn't inc counter
    renew_sid = Rack::Session::XFile.new(renew_session)   # increments counter

    res = Rack::MockRequest.new(inc_cntr).get('/')
    cookie = res['Set-Cookie']
    res.body.should.equal counter(1)

    res = Rack::MockRequest.new(read_sid).get('/', set_sid(cookie))
    cookie = res['Set-Cookie'] if res['Set-Cookie']
    res.body.length.should.equal session_chars
    old_sid = res.body

    res = Rack::MockRequest.new(renew_sid).get('/', set_sid(cookie))
    cookie = res['Set-Cookie'] if res['Set-Cookie']
    res.body.should.equal counter(2)

    res = Rack::MockRequest.new(read_sid).get('/', set_sid(cookie))
    cookie = res['Set-Cookie'] if res['Set-Cookie']
    res.body.length.should.equal session_chars
    new_sid = res.body

    # the session id should have changed, ...
    new_sid.should.not.equal old_sid

    # but the session data should persist
    res = Rack::MockRequest.new(inc_cntr).get('/', set_sid(cookie))
    res.body.should.equal counter(3)

    # and there should only be one xfile session
    inc_cntr.session_count.should.equal 1
  end

  it 'returns no cookie and deletes xfile with :drop option set' do
    app = Rack::Session::XFile.new(incrementor)
    req = Rack::MockRequest.new(app)
    drop = Rack::Utils::Context.new(app, drop_session)
    dreq = Rack::MockRequest.new(drop)

    res = req.get('/')
    original_session = res['Set-Cookie'][TheSession]
    res.body.should.equal counter(1)
    app.session_count.should.equal 1

    res = req.get('/', set_sid(original_session))
    res.body.should.equal counter(2)
    app.session_count.should.equal 1

    res = dreq.get('/', set_sid(original_session))
    res['Set-Cookie'].should.be.nil
    res.body.should.equal counter(3)
    app.session_count.should.be.zero

    res = req.get('/', set_sid(original_session))
    res['Set-Cookie'][TheSession].should.not.equal original_session
    res.body.should.equal counter(1)
    app.session_count.should.equal 1
  end

  it 'returns no cookie and updates xfile with :defer option set' do
    app   = Rack::Session::XFile.new(incrementor)
    exp   = Rack::Session::XFile.new(incrementor, expire_after:3600)
    exdfr = Rack::Session::XFile.new(incrementor, expire_after:3600, defer:true)
    defer = Rack::Session::XFile.new(defer_session)

    res = Rack::MockRequest.new(defer).get('/')
    res['Set-Cookie'].should.be.nil     # normally, new sessions set cookies
    res.body.should.equal counter(1)    # this xfile is updated on the server,
    app.session_count.should.equal 1    # but is not referenced by any client

    res = Rack::MockRequest.new(app).get('/')
    cookie = res['Set-Cookie']
    cookie.should.not.be.nil            # because new session created
    res.body.should.equal counter(1)
    app.session_count.should.equal 2

    res = Rack::MockRequest.new(app).get('/', set_sid(cookie))
    res['Set-Cookie'].should.be.nil     # because sid did not change
    res.body.should.equal counter(2)
    app.session_count.should.equal 2

    res = Rack::MockRequest.new(exp).get('/', set_sid(cookie))
    cookie = res['Set-Cookie']
    cookie.should.not.be.nil            # because :expire_after forced update
    res.body.should.equal counter(3)
    app.session_count.should.equal 2

    res = Rack::MockRequest.new(exdfr).get('/', set_sid(cookie))
    res['Set-Cookie'].should.be.nil     # because :defer overrides :expire_after
    res.body.should.equal counter(4)
    app.session_count.should.equal 2

    res = Rack::MockRequest.new(app).get('/', set_sid(cookie))
    res['Set-Cookie'].should.be.nil     # because sid did not change,
    res.body.should.equal counter(5)    # but keeps state even after :defer
    app.session_count.should.equal 2
  end

  it 'returns no cookie and updates xfile with :defer option set in context' do
    app   = Rack::Session::XFile.new(incrementor)
    defer = Rack::Utils::Context.new(app, defer_session)

    res = Rack::MockRequest.new(defer).get('/')
    res['Set-Cookie'].should.be.nil
    res.body.should.equal counter(1)
    app.session_count.should.equal 1
  end

  it 'returns no cookie and skips session processing for filtered requests' do
    human_user_agent = 'Mozilla/5.0 CoolSurf/100.50'
    robot_user_agent = 'BigSpider v1.0'
    robot_filter = /(bot|spider|crawler)/i
    app = Rack::Session::XFile.new(incrementor, user_agent_filter: robot_filter)
    req = Rack::MockRequest.new(app)

    res = req.get('/', 'HTTP_USER_AGENT' => robot_user_agent)
    res['Set-Cookie'].should.be.nil
    app.session_count.should.be.zero

    res = req.get('/', 'HTTP_USER_AGENT' => human_user_agent)
    res['Set-Cookie'].should.not.be.nil
    app.session_count.should.equal 1
  end

  it 'returns no cookie if sid is unchanged' do
    app = Rack::Session::XFile.new(incrementor)
    req = Rack::MockRequest.new(app)

    cookie = req.get('/')['Set-Cookie']
    cookie.should.match TheSession

    res = req.get('/', set_sid(cookie))
    res.body.should.equal counter(2)
    res['Set-Cookie'].should.be.nil

    res = req.get('/', set_sid(cookie))
    res.body.should.equal counter(3)
    res['Set-Cookie'].should.be.nil
  end

  it 'returns no cookie if session is not read or written' do
    app = Rack::Session::XFile.new(noop)
    res = Rack::MockRequest.new(app).get('/')
    res['Set-Cookie'].should.be.nil
    app.session_count.should.be.zero
  end

  it 'returns no cookie if previous session is not read or written' do
    app = Rack::Session::XFile.new(noop)
    res = Rack::MockRequest.new(app).get('/', set_sid(create_session))
    res['Set-Cookie'].should.be.nil
    app.session_count.should.equal 1
  end

  it 'returns no cookie if session is read but not written' do
    app = Rack::Session::XFile.new(read_session)
    res = Rack::MockRequest.new(app).get('/')
    res['Set-Cookie'].should.be.nil
    app.session_count.should.be.zero
  end

  it 'returns no cookie if previous session is read but not written' do
    app = Rack::Session::XFile.new(read_session)
    res = Rack::MockRequest.new(app).get('/', set_sid(create_session))
    res['Set-Cookie'].should.be.nil
    app.session_count.should.equal 1
  end

  it 'returns no cookie if session is not written and was not '\
     'previously created, regardless of the :expire_after option' do
    app = Rack::Session::XFile.new(read_session, expire_after: 3600)
    res = Rack::MockRequest.new(app).get('/')
    res['Set-Cookie'].should.be.nil
    app.session_count.should.be.zero
  end

  it 'returns cookie if previous session is not read or written '\
     'but :expire_after option is set' do
    app = Rack::Session::XFile.new(noop, expire_after: 3600)
    res = Rack::MockRequest.new(app).get('/', set_sid(create_session))
    res['Set-Cookie'].should.not.be.nil
    app.session_count.should.equal 1
  end

  FileUtils.rm_rf(spec_tmp_dir)
end
